@@ -17,7 +17,7 @@ JEDI_VERSION = jedi.__version__
 
 # Eol chars accepted by the LSP protocol
 # the ordering affects performance
-EOL_CHARS = ['\r\n', '\r', '\n']
+EOL_CHARS = ["\r\n", "\r", "\n"]
 EOL_REGEX = re.compile(f'({"|".join(EOL_CHARS)})')
 
 log = logging.getLogger(__name__)
@@ -25,6 +25,7 @@ log = logging.getLogger(__name__)
 
 def debounce(interval_s, keyed_by=None):
     """Debounce calls to this function until interval_s seconds have passed."""
+
     def wrapper(func):
         timers = {}
         lock = threading.Lock()
@@ -48,7 +49,9 @@ def debounce(interval_s, keyed_by=None):
                 timer = threading.Timer(interval_s, run)
                 timers[key] = timer
                 timer.start()
+
         return debounced
+
     return wrapper
 
 
@@ -78,7 +81,9 @@ def find_parents(root, path, names):
     # Search each of /a/b/c, /a/b, /a
     while dirs:
         search_dir = os.path.join(*dirs)
-        existing = list(filter(os.path.exists, [os.path.join(search_dir, n) for n in names]))
+        existing = list(
+            filter(os.path.exists, [os.path.join(search_dir, n) for n in names])
+        )
         if existing:
             return existing
         dirs.pop()
@@ -92,11 +97,11 @@ def path_to_dot_name(path):
     directory = os.path.dirname(path)
     module_name, _ = os.path.splitext(os.path.basename(path))
     full_name = [module_name]
-    while os.path.exists(os.path.join(directory, '__init__.py')):
+    while os.path.exists(os.path.join(directory, "__init__.py")):
         this_directory = os.path.basename(directory)
         directory = os.path.dirname(directory)
         full_name = [this_directory] + full_name
-    return '.'.join(full_name)
+    return ".".join(full_name)
 
 
 def match_uri_to_workspace(uri, workspaces):
@@ -128,6 +133,7 @@ def merge_dicts(dict_a, dict_b):
 
     If override_nones is True, then
     """
+
     def _merge_dicts_(a, b):
         for key in set(a.keys()).union(b.keys()):
             if key in a and key in b:
@@ -143,6 +149,7 @@ def merge_dicts(dict_a, dict_b):
                 yield (key, a[key])
             elif b[key] is not None:
                 yield (key, b[key])
+
     return dict(_merge_dicts_(dict_a, dict_b))
 
 
@@ -150,8 +157,8 @@ def escape_plain_text(contents: str) -> str:
     """
     Format plain text to display nicely in environments which do not respect whitespaces.
     """
-    contents = contents.replace('\t', '\u00A0' * 4)
-    contents = contents.replace('  ', '\u00A0' * 2)
+    contents = contents.replace("\t", "\u00A0" * 4)
+    contents = contents.replace("  ", "\u00A0" * 2)
     return contents
 
 
@@ -160,17 +167,17 @@ def escape_markdown(contents: str) -> str:
     Format plain text to display nicely in Markdown environment.
     """
     # escape markdown syntax
-    contents = re.sub(r'([\\*_#[\]])', r'\\\1', contents)
+    contents = re.sub(r"([\\*_#[\]])", r"\\\1", contents)
     # preserve white space characters
     contents = escape_plain_text(contents)
     return contents
 
 
 def wrap_signature(signature):
-    return '```python\n' + signature + '\n```\n'
+    return "```python\n" + signature + "\n```\n"
 
 
-SERVER_SUPPORTED_MARKUP_KINDS = {'markdown', 'plaintext'}
+SERVER_SUPPORTED_MARKUP_KINDS = {"markdown", "plaintext"}
 
 
 def choose_markup_kind(client_supported_markup_kinds: List[str]):
@@ -181,10 +188,12 @@ def choose_markup_kind(client_supported_markup_kinds: List[str]):
     for kind in client_supported_markup_kinds:
         if kind in SERVER_SUPPORTED_MARKUP_KINDS:
             return kind
-    return 'markdown'
+    return "markdown"
 
 
-def format_docstring(contents: str, markup_kind: str, signatures: Optional[List[str]] = None):
+def format_docstring(
+    contents: str, markup_kind: str, signatures: Optional[List[str]] = None
+):
     """Transform the provided docstring into a MarkupContent object.
 
     If `markup_kind` is 'markdown' the docstring will get converted to
@@ -195,33 +204,24 @@ def format_docstring(contents: str, markup_kind: str, signatures: Optional[List[
     to the provided contents of the docstring if given.
     """
     if not isinstance(contents, str):
-        contents = ''
+        contents = ""
 
-    if markup_kind == 'markdown':
+    if markup_kind == "markdown":
         try:
             value = docstring_to_markdown.convert(contents)
-            return {
-                'kind': 'markdown',
-                'value': value
-            }
+            return {"kind": "markdown", "value": value}
         except docstring_to_markdown.UnknownFormatError:
             # try to escape the Markdown syntax instead:
             value = escape_markdown(contents)
 
         if signatures:
-            value = wrap_signature('\n'.join(signatures)) + '\n\n' + value
+            value = wrap_signature("\n".join(signatures)) + "\n\n" + value
 
-        return {
-            'kind': 'markdown',
-            'value': value
-        }
+        return {"kind": "markdown", "value": value}
     value = contents
     if signatures:
-        value = '\n'.join(signatures) + '\n\n' + value
-    return {
-        'kind': 'plaintext',
-        'value': escape_plain_text(value)
-    }
+        value = "\n".join(signatures) + "\n\n" + value
+    return {"kind": "plaintext", "value": escape_plain_text(value)}
 
 
 def clip_column(column, lines, line_number):
@@ -230,7 +230,9 @@ def clip_column(column, lines, line_number):
 
     https://microsoft.github.io/language-server-protocol/specification#position
     """
-    max_column = len(lines[line_number].rstrip('\r\n')) if len(lines) > line_number else 0
+    max_column = (
+        len(lines[line_number].rstrip("\r\n")) if len(lines) > line_number else 0
+    )
     return min(column, max_column)
 
 
@@ -242,14 +244,16 @@ def position_to_jedi_linecolumn(document, position):
     """
     code_position = {}
     if position:
-        code_position = {'line': position['line'] + 1,
-                         'column': clip_column(position['character'],
-                                               document.lines,
-                                               position['line'])}
+        code_position = {
+            "line": position["line"] + 1,
+            "column": clip_column(
+                position["character"], document.lines, position["line"]
+            ),
+        }
     return code_position
 
 
-if os.name == 'nt':
+if os.name == "nt":
     import ctypes
 
     kernel32 = ctypes.windll.kernel32
